#include<Windows.h>
#include <string.h>
#include <stdio.h>


#define SOH 0x01
#define STX 0x02
#define EOT 0x04
#define ACK 0x06
#define NAK 0x15
#define CAN 0x18
#define CRC 0x43
#define ERROR_MAX 30

const char *filename = "RUI_4.0.0_RAK3172-E.bin";

extern void serial_write(const char* data, int len); 
extern int serial_read(char *data, int wait_ms); 

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t; 


void send_AT(const char *cmd)
{
    int len = strlen(cmd); 
    serial_write(cmd, len);        
}

void enter_boot_mode(void)
{
	char data[256] ;
	int len = 0; 

    send_AT("at+boot\n");   
    
    len = serial_read(data, 1000); 
}

bool check_boot_mode(void)
{
    char data[64];

    send_AT("at+\r\n");

    serial_read(data, 1000); 

    if (strstr(data, "AT not support") != NULL) {
        printf("[dbg]:Device is in boot mode\n");
        return true; 
    }
    else
        if (strstr(data, "OK") != NULL || 
            strstr(data, "AT_ERROR") !=NULL ||
            strstr(data, "AT_COMMAND_NOT_FOUND") != NULL)
            printf("[dbg]:Device is not in boot mode \n");
    return false;
}


bool enter_dfu_mode(void)
{
    char data[64];

    send_AT("at+update\r\n") ;
    serial_read(data, 1000); 

    if (strstr(data, "Start firmware upgrade") != NULL) {
        printf("[dbg]:Device enter dfu mode\n");
        return true; 
    }
    else
        printf("[dbg]:enter dfu mode fail \n");
    return false;

} 


//-------------------------------------------
// CRC function 
//-------------------------------------------
uint16_t crc_update(uint16_t crc_in, int incr)
{
    uint16_t xr = crc_in >> 15;
    uint16_t out = crc_in << 1;

    if (incr)
        out++;

    if (xr)
        out ^= 0x1021;

    return out;
}

uint16_t crc16(const uint8_t *data, uint16_t size)
{
    uint16_t crc, i;

    for (crc = 0; size > 0; size--, data++)
        for (i = 0x80; i; i >>= 1)
            crc = crc_update(crc, *data & i);

    for (i = 0; i < 16; i++)
        crc = crc_update(crc, 0);

    return crc;
}

uint16_t swap16(uint16_t in)
{
  return (in >> 8) | ((in & 0xff) << 8);
}


//-------------------------------------------

struct ymodem_block {
    uint8_t header[3];
    uint8_t data[1024];
    uint8_t crc[2];
};

//--------------------------------
// Ymodem upgrade main function 
//--------------------------------

bool ymodem_upgrade()
{
    uint8_t ret; 
    char data[64]; 
    uint8_t reading_char;
    uint8_t error_count = 0;
    uint8_t success_count = 0;
    bool cancel = false;
    uint8_t sequence = 0; 
    struct ymodem_block sending_buf; 
    FILE *fp ;
    int filesize;    
    int read_size; 

    //
    // Step#1 : Starting Ymodem sync up with peer device
    //
    while(1)
    {
        ret = serial_read(data, 100);
        if(ret)
        {
            reading_char = data[0];        
            if(reading_char == NAK)
            {
                printf("[S] STATE: Received checksum request (NAK)\n");
                printf("[S] STATE: Checksum mode applied\n");
                break;
            }
            else if(reading_char == CAN)
            {
                printf("received CAN\n"); 
                if(cancel)
                {
                    printf("[S] STATE: Transmission cancelled (Received 2 CANs at mode request)\n");
                    return false;  
                }
                else
                {
                    printf("[S] STATE: Ready for transmission cancellation\n");
                    cancel = 1;
                }
                break; 
            }        
            else if(reading_char == CRC)
            {    
                printf("[S] STATE: Received CRC request (C/CRC)\n");
                printf("[S] STATE: 16-bit CRC mode applied\n");
                break;
            }        
            else if(reading_char == EOT)
            {    
                printf("[S] STATE: Transmission cancelled (Received EOT at mode request)\n");
                return false; 
            }
            else 
                printf("[S] ERROR: Expected NAK, CRC, EOT or CAN but got %r\n", reading_char);

            if(error_count++ > ERROR_MAX)
            {
                printf("[S] ERROR: error_count reached %d, aborting...\n", error_count);
                return false;
            }
        }
    }
    
    printf("[S] STATE: Preparing info block\n");
    
    //
    // Step#2 : Prepare Information Block
    //

    fopen_s(&fp, filename,"rb");
    if(fp == NULL)
    {
        printf("Firmware Error \n!");   
        return false;              
    }
    /* get file size */
    fseek(fp, 0L, SEEK_END);
    filesize = ftell(fp);
    fclose(fp); 

    memset((void *)&sending_buf,0,sizeof(struct ymodem_block));
    sending_buf.header[0] = STX;
    sending_buf.header[1] = sequence;
    sending_buf.header[2] = 0xff - sequence;
    
    /* [required] Name : string */
    int len = strlen(filename);
    memcpy(sending_buf.data, filename,len);
    
    /* [Optional] Length : string*/
    sprintf((char *)(sending_buf.data+len+1),"%d",filesize);
    int crc = crc16(sending_buf.data, 1024); 
    sending_buf.crc[0] = crc >> 8;
    sending_buf.crc[1] = (crc & 0xff);

    /* send the 1st ymodem information block to peer device */
    serial_write((const char *)&sending_buf, sizeof(struct ymodem_block));
    
    printf("[S] TRANSMISSION: info block sent\n");
    error_count = 0;

    /* wait ACK response */
    while(1)
    {
        ret = serial_read(data, 10);
        if(ret)
        {
            reading_char = data[0];                
            if(reading_char == ACK)
            {    
                printf("[S] RECEIVE: ACK received \n");
                break;
            }
            printf("Expected ACK but got %u(%c) for info block \n",reading_char, reading_char);
            if(error_count++ > ERROR_MAX)
            {
                printf("error count reach MAX, abort \n");
                return false;
            }
        }
    }
    printf("[S] STATE: Waiting the mode request...\n");
    error_count = 0;
    cancel = false;
    sequence = 1 ; 

    while(1)
    {
        ret = serial_read(data, 500);
        if(ret)
        {
            reading_char = data[0];  
            if(reading_char == NAK)
            {       
                printf("[S] STATE: Received checksum request (NAK)\n");
                printf("[S] STATE: Checksum mode applied\n");
                break;
            }    
            else if(reading_char == CRC)
            {
                printf("[S] STATE: Received CRC request (C/CRC)\n"); 
                printf("[S] STATE: 16-bit CRC mode applied\n");
                break;
            }
            else if(reading_char == CAN)
            {    
                printf("received CAN\n");
                if (cancel)
                {
                    printf("[S] TRANSMISSION: Cancelled (Received 2 CANs at mode request)\n");
                    return false; 
                }
                else
                {
                    printf("[S] STATE: Ready for transmission cancellation\n");
                    cancel = 1;
                }
            }
            else if (reading_char == EOT)
            {
                printf("[S] TRANSMISSION: Cancelled (Received EOT at mode request)\n");
                return false;
            }
            else
                printf("[S] ERROR: Expected NAK, CRC, EOT or CAN but got %r\n", reading_char); 
                        
            if (++error_count > ERROR_MAX)
            {    
                printf("[S] ERROR: error_count reached %d, aborting...\n", ERROR_MAX);
                return false; 
            }
        }
    }    
    
    //
    // Step#3 : Starting File of Transmission 
    //

    
    fp = fopen(filename, "rb");
    if (fp == NULL)
    {
        printf("Firmware Error \n!");
        return false;
    }
    
    while (1)
    {
        /* zero buffer */
        memset((void*)&sending_buf, 0, sizeof(struct ymodem_block));
        read_size = fread(sending_buf.data, 1, 1024, fp);

        if (read_size == 0)
        {
            printf("[S] TRANSMISSION: Reached EOF\n");
            break;
        }
        else if (read_size < 1024)
        {
            memset((void*)&sending_buf.data[read_size], 0x1a, 1024-read_size);
        }

        sending_buf.header[0] = STX;
        sending_buf.header[1] = sequence;
        sending_buf.header[2] = 0xff - sequence;

        crc = crc16(sending_buf.data, 1024); 
        sending_buf.crc[0] = crc >> 8;
        sending_buf.crc[1] = (crc & 0xff);

        while(1) 
        {
            serial_write((const char *)&sending_buf, sizeof(struct ymodem_block));      
            printf("[S] TRANSMISSION: block {%d} (seq={%d}) sent\n", success_count, sequence); 
            ret = serial_read(data, 10);
            if(ret)
            {
                reading_char = data[0];
                if (reading_char == ACK)
                {
                    success_count++;
                    error_count = 0;
                    break;
                }
                else if (reading_char == CAN)
                {
                    if (cancel)
                    {
                        printf("[S] TRANSMISSION: Cancelled (Received 2 CANs at transmission)\n");
                        return false;
                    }
                    else
                    {
                        printf("[S] STATE: Ready for transmission cancellation\n");
                        cancel = 1;
                    }
                }
                printf("[S] ERROR: Expected ACK but got %X for block %d\n", reading_char, sequence);
                error_count += 1;

                if (error_count > ERROR_MAX)
                {
                    printf("[S] ERROR: NAK received {} times, aborting...\n", error_count);
                    return false;
                }
            }    
        }
        sequence = (sequence + 1) % 0x100;   
    }
    
    // 
    // Step#4 : End of Transmission 
    //
    while(1)
    {
        const char eot = EOT; 
        serial_write(&eot, sizeof(eot));      
        printf("[S] TRANSMISSION: EOT sent and awaiting ACK\n");
        ret = serial_read(data, 10);
        if(ret)
        {
            reading_char = data[0];
            if(reading_char == ACK)
                break;
            else
            {
                printf("[S] ERROR: Expected ACK but got %X \n", reading_char);
                if (++error_count > ERROR_MAX)
                {
                    printf("[S] WARN: EOT was not ACKd, aborting transfer...\n");
                    return false;
                }
            }
        }   
    }
    printf("[S] TRANSMISSION: Finished (ACK received)\n");

    while(1)
    { 
        /* zero buffer */
        memset((void *)&sending_buf,0,sizeof(struct ymodem_block));
        sending_buf.header[0] = STX;
        sending_buf.header[1] = 0x00;
        sending_buf.header[2] = 0xff;
            
        int crc = crc16(sending_buf.data, 1024); 
        sending_buf.crc[0] = crc >> 8;
        sending_buf.crc[1] = (crc & 0xff);

        serial_write((const char *)&sending_buf, sizeof(struct ymodem_block));
        
        printf("[S] TRANSMISSION: SEND NULL\n");
        ret = serial_read(data, 10);
        if(ret)
        {
            reading_char = data[0];
            if(reading_char == ACK)
                break;                
            else
            {    
                printf("[S] ERROR: Expected ACK but got %02X \n", reading_char); 
                
                if (++error_count > ERROR_MAX)
                {
                    printf("[S] WARN: EOT was not ACKd, aborting transfer... \n"); 
                    return false; 
                }
            }
        }
    }
    return true; 
}
